﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2012 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System.Collections.Generic;
using System.Linq;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Optimization.Operators.LCS;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.Encodings.DecisionList {
  [Item("SmartDecisionListCreator", "Description missing")]
  [StorableClass]
  public class SmartDecisionListCreator : DecisionListCreator {

    [StorableConstructor]
    protected SmartDecisionListCreator(bool deserializing) : base(deserializing) { }
    protected SmartDecisionListCreator(SmartDecisionListCreator original, Cloner cloner)
      : base(original, cloner) {
    }
    public SmartDecisionListCreator()
      : base() {
    }
    public override IDeepCloneable Clone(Cloner cloner) {
      return new SmartDecisionListCreator(this, cloner);
    }

    protected override DecisionList Create(IRandom random, Rule sampleRule, int initialNumberOfRules, double oneProbability, ItemCollection<IDiscretizer> discretizers, IAction niche) {
      var problemData = ProblemDataParameter.ActualValue;

      List<Rule> rules = new List<Rule>();
      for (int i = 0; i < initialNumberOfRules; i++) {
        rules.Add(SmartRuleInitialization(problemData, random, sampleRule, initialNumberOfRules, oneProbability, discretizers, niche));
      }
      if (niche != null) {
        return new DecisionList(rules, niche);
      } else {
        return new DecisionList(rules);
      }
    }

    public static Rule SmartRuleInitialization(IDecisionListClassificationProblemData problemData, IRandom random,
                                               Rule sampleRule, int initialNumberOfRules, double oneProbability,
                                               ItemCollection<IDiscretizer> discretizers, IAction niche = null) {
      IList<int> trainingIndicesOfOtherNiches = null;
      if (niche != null) {
        trainingIndicesOfOtherNiches =
          problemData.TrainingIndices.Where(x => !problemData.FetchAction(x).SameNiche(niche)).ToList();
      } else {
        trainingIndicesOfOtherNiches = problemData.TrainingIndices.ToList();
      }

      var newRule = (Rule)sampleRule.Clone();
      newRule.Randomize(random, oneProbability, discretizers);

      int sample = random.Next(trainingIndicesOfOtherNiches.Count);
      var input = problemData.FetchInput(trainingIndicesOfOtherNiches[sample]);
      var action = problemData.FetchAction(trainingIndicesOfOtherNiches[sample]);

      newRule.SetToMatchInput(input);
      newRule.Action.SetTo(action);
      return newRule;
    }
  }
}
